#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

# Collection of astronomical procedures. Used by sourcing from circ.tcl,
# tlsp.tcl.
#
# Astronomical computations here are made using formulii given in two books:
# 1. See Astronomija s personal'nym compjuterom, Montenbruk, Pfleger, Moskva, 
# "Mir", 1993, abbreviated here as APC,
# 2. Astronomicheskie formuly dlja kal'kuljatora, Mees, Moskva, "Mir", 1988,
# abbreviated here as AFK.
#
# Compute Julian day for given date: see AFK, p.27
#
# @param date date in format "YYYY MM DD[.dd]"
# @return Julian day (conventional, i.e. NOT with 2400000 subtracted)
#
# NOTE that if day part in date is integer, the fractional part of the result 
# will be 0.5.
#
proc jd date {
	set l [split $date]
	# below we add ".0" in the end of numbers to avoid treatment them as octal
	# integers by Tcl, if the start with "0" (e.g. 08 is illegal octal, 08.0==8)
	set yy [lindex $l 0].0;	set mm [lindex $l 1].0; set dd [lindex $l 2];
	if {[string match 0* $dd] && ![string match *.* $dd]} {set dd "$dd.0"}
	if {$mm<=2} {set mm [expr $mm+12]; set yy [expr $yy-1]}
	if {[expr $yy+$mm/100+$dd/10000]>=1582.1015} {
		set a [expr int($yy/100.)]; 
		set b [expr 2. - $a + int($a/4.)]
	} else { set b 0}
	set jd [expr int(365.25*$yy - 0.75*($yy<0)) + int(30.6001 * ($mm+1)) + $dd\
		+ 1720994.5 + $b]
	return $jd
}

#
# Compute greenwich stellar time for 0hUT of given JD (AFK p.37)
# @param jd Julian date
# @return st0 in hours [0,24)
#
# NOTE that jd may contain the fractions of day (i.e. frac(jd)!=0.5) but it is
# truncated to the jd of the beginning of the same date before calculations.
#  Example: [st0 2450000.6789] == [st0 2450000.5], but
# [st0 2450000.4567] == [st0 2449999.5].
#
# The reason to refrain from calculations of JD for the arbitrary moment of the
# day is the loss of significance of the least significant digits in the result.
# In order to compute stellar time for arbitrary moment of time and for non-zero
# longitude - use st{}
#
proc st0 jd {
	# JD for 0hUT:
	set jd0 [expr int($jd+0.5)-0.5]
	set t [expr ($jd0 - 2415020.0)/36525]
		# Greenwich time for 0hUT in fractional hours:
		#set theta0 [expr 6.6460656 + 2400.051262 * $t + 0.00002581 * $t*$t]
	# Greenwich time for 0hUT in fractional days:
	set theta0 [expr 0.276919398 + 100.0021359 * $t + 0.000001075 *$t*$t]
	return [expr 24. * [frac $theta0]]
}

#
# Get local stellar time
# @param date date in format accepted by jd{}
# @param hours time in hours
# @param long longitude in hours
# @return Stellar time in hours [0,24)
#
proc st {date hours long} {
	set st [expr [st0 [jd $date]] + $hours*1.002737908 + $long]
	return [expr fmod($st,24)+($st<0)*24]
}


# For mini_sun:
# @return  fractional part of x counted from the largest integer less than x
proc frac x {
	return [expr $x - floor($x)] 
}
#
# Compute approximate equatorial coordinates of Sun from JD with prec.~1arcmin
# @param jd Julian day containing the fraction of date:
# 	use [expr [jd <date>] + [xms2x <time>]/24.] as an argument.
# @return [list sunra sundec] in hours and degrees, respectively
# See APC, p.48
#
proc mini_sun jd {
	# Time in julian centuries:
	set t [expr ($jd-2451545.0)/36525.0]
	set p2 6.283185307;	set coseps 0.91748; set sineps 0.39778
	set m [expr $p2 * [frac [expr 0.993133 + 99.997361 * $t]]]
	set dl [expr 6893.0 * sin($m) + 72.0*sin(2*$m)]
	set l [expr $p2 * [frac [expr 0.7859453 + $m/$p2 + (6191.2*$t+$dl)/1296.e3]]]
	set sl [expr sin($l)]
	set x [expr cos($l)]; set y [expr $coseps*$sl]; set z [expr $sineps*$sl]
	set rho [expr sqrt(1.0-$z*$z)]
	set sundec [expr (360.0/$p2)*atan($z/$rho)]
	set sunra [expr (48./$p2)*atan($y/($x+$rho))]
	if {$sunra<0} {set sunra [expr $sunra+24.0]}
	return [list $sunra $sundec]
}
#
# Compute approximate equatorial coordinates of Moon from JD with prec.~5arcmin
# @param jd Julian day containing the fraction of date:
# 	use [expr [jd <date>] + [xms2x <time>]/24.] as an argument.
# @return [list moonra moondec] in hours and degrees, respectively
# See APC, p.47
#
proc mini_moon jd {
	# Time in julian centuries:
	set t [expr ($jd-2451545.0)/36525.0]
	set p2 6.283185307;	set coseps 0.91748; set sineps 0.39778
	set arc 206264.8062
	
	set l0 [frac [expr 0.606433 + 1336.855225 * $t]]
	set l  [expr $p2 * [frac [expr 0.374897 + 1325.552410 * $t]]]
	set ls [expr $p2 * [frac [expr 0.993133 +   99.997361 * $t]]]
	set d  [expr $p2 * [frac [expr 0.827361 + 1236.853086 * $t]]]
	set f  [expr $p2 * [frac [expr 0.259086 + 1342.227825 * $t]]]
	
	set dl [expr 22640*sin($l) - 4586*sin($l-2*$d) + 2370*sin(2*$d) + \
		769*sin(2*$l) - 668*sin($ls) - 412*sin(2*$f) - \
		212*sin(2*$l-2*$d) - 206*sin($l+$ls-2*$d) + 192*sin($l+2*$d)- \
		165*sin($ls-2*$d) - 125*sin($d) - 110*sin($l+$ls) + \
		148*sin($l-$ls) - 55*sin(2*$f-2*$d)]
	set s [expr $f + ($dl + 412*sin(2*$f) + 541*sin($ls))/$arc]
	set h [expr $f-2*$d]
	set n [expr -526*sin($h) + 44*sin($l+$h) - 31*sin($h-$l) - 23*sin($ls+$h) \
		+ 11*sin($h-$ls) - 25*sin($f-2*$l) + 21*sin($f-$l)]
		
	# ecliptical coordinates in radians:
	set l_moon [expr $p2 * [frac [expr $l0 + $dl/1296e3]]]
	set b_moon [expr (18520.0*sin($s) + $n)/$arc]
	# puts "l=[expr $l_moon*360/$p2] deg, b=[expr $b_moon*360/$p2] deg"
	
	set cb [expr cos($b_moon)]
	set x [expr $cb*cos($l_moon)]
	set v [expr $cb*sin($l_moon)]
	set w [expr sin($b_moon)]
	set y [expr $coseps*$v-$sineps*$w]
	set z [expr $sineps*$v+$coseps*$w]
	set rho [expr sqrt(1.0-$z*$z)]
	set moondec [expr (360.0/$p2)*atan($z/$rho)]
	set moonra [expr (48./$p2)*atan($y/($x+$rho))]

	if {$moonra<0} {set moonra [expr $moonra+24.0]}
	return [list $moonra $moondec]
}

#
# Convert equatorial coordinates HA,DEC in horizontal A,h (AFK p.41)
# @param ha hour angle in hours
# @param dec declination in degrees
# @param lat latitude in degrees
# @return [list azimuth altitude] in degrees
#
proc hadec2azalt {ha dec lat} {
	set pi 3.14159265358979323
	set ha [expr $ha*$pi/12.]
	set dec [expr $dec*$pi/180.]
	set fi [expr $lat*$pi/180.]
	set sinh [expr sin($fi)*sin($dec) + cos($fi)*cos($dec)*cos($ha)]
	set a [expr atan2(sin($ha), cos($ha)*sin($fi)-tan($dec)*cos($fi))*180/$pi]
	set h [expr asin($sinh)*180/$pi]
	return [list $a $h]
}

#
# Convert degrees in hh mm ss:
#
proc deg2hms deg {
	if {$deg<0 || $deg>=360} {
		return [x2xms [expr 24.0 * [frac [expr $deg/360.0] ]]]
	} else { 
		return [x2xms [expr $deg/15.]]
	}
}
#
# Convert degrees or hours in dd mm ss.ss or hh mm ss.ss
#
# No correction for the range is applied. Seconds are given with 0.003
# precision (others are not significant for sure).
# For negative values the "-" sign is put as a leading symbol in a resulting 
# string, even if the resulting degree is zero: e.g. [x2xms -0.5]=="-0 30 0".
#
proc x2xms x {
	if {$x<0} {set sig "-"} else {set sig ""}
	set x [expr abs($x)]
	set d [expr int($x)]
	set m [expr int(($x-$d)*60.0)]
	set s [expr ($x-$d-$m/60.)*3600.0]
	return [list $sig$d $m [format %.3f $s]]
}
#
# Convert xx mm ss into fractional x (degrees or hours)
#
# NOTES:
# 1. In argument representation, no leading spaces are allowed and the hours,
# minutes and seconds must be separated by the SINGLE space character.
#
# 2. Arguments HH and MM must be integers (i.e. do not contain decimal point)
#
# 3. For negative values in range either of xx, mm or ss MUST contain the
# minus-sign to obtain the correct negative result (e.g.: "-00 12 34",
# "00 -12 34" and "00 -12 -34" are equivalent arguments).
#
# @param xxmmss string in format "XX MM SS" where XX are integer hours or
# 		degrees, MM are integer minutes and SS are integer or fractional seconds
# @return the fractional angle (or hour) x.
#
proc xms2x xxmmss {
	set l [split $xxmmss]
	# below we add ".0" in the end of numbers to avoid treatment them as octal
	# integers by Tcl, if the start with "0" (e.g. 08 is illegal octal, 08.0==8)
	set xx [lindex $l 0].0;	set mm [lindex $l 1].0; set ss [lindex $l 2];
	if {[string match 0* $ss] && ![string match *.* $ss]} {set ss "$ss.0"}
	set r [expr abs($xx) + abs($mm)/60. + abs($ss)/3600.]
	set sig [expr 1. - 2*($xx<0 || $mm<0 || $ss<0 || [string match -* $xx])]
	return [expr $sig*$r]
}
