#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

# Observation circumstance computer prototype.
# Current version (see "ident") computes real hsun using the local machine time
# (precise time should be set! Use XNTPD for example)
# and the longitude and latitude supplied by SV circumstance monitor scenario 
# It still supports the 2-minutes day simulation if set in test-mode (see
# below), BUT uses real settings of geographic coordinates, etc. Starting point
# of simulation is conditions determined for the current moment of time, but the
# test-mode stimulates the time to go 360 times faster, i.e. the simulated day 
# lasts only 4 minutes.
#
# Astronomy formula used here are sourced from the separate script:
source astro.tcl
#
# Apart from standard general commands (described in commands.tex in section
# General commands) the following commands are supported:
#
# SET LONGITUDE="+-DD MM SS" LATITUDE="+-HH MM SS" 
# Note, that in CURRENT VERSION of CIRC the agreement about the geographic 
# longitude is accepted, that longitude is POSITIVE TO EAST which is in
# agreement with current international standards. Latitude is positive for Nord.
# I.e. Moscow has a positive longitude, while Cerro Tololo - negative)
# 		to override wrong defaults.
# SET MODE=TEST|REAL - to switch between simulated 2-minutes winter day and real
# computations
# GET UT - returns current UT as "YYYY MM DD hh mm ss" with 1sec precision
# GET ST - return current local stellar time as "hh mm ss.sss"
# GET STH - return current local stellar time as floating-point hour hh.hhhh
# GET JD - returns current Julian Day
# GET HSUN - returns current true (not apparent!) altitude of Sun above horizon 
# in degrees. (Note: HSUN=0 corresponds to apparent altitude of the CENTER of
# the solar disk of +34 arcmin. Sunset corresponds to HSUN=-0.8 degree.)
# GET SUNRA SUNDEC - return equatorial coords of Sun (hours, degrees) (+-1')
# GET MOONRA MOONDEC - return equatorial coords of Moon (hours, degrees) (+-5')
#

source compcmd.tcl

#
# Port number:
set port 1234
# Identification:
set ident "circ.tcl v0.1"

# Parameters allowed for setting without checking (appended in objm.tcl):
set par_set ""

# Default geographic coordinates: Moscow
# in hours minutes seconds:
# ATTENTION!!!!!! Longitude is positive to the EAST from Greenwich!!!!!!!!!
set longitude "2 30 48"
set long [xms2x $longitude]
# in degree minites seconds:
set latitude "55 45 00"
set lat [xms2x $latitude]

# Mode: test/real
set realmode true

# Test day is shorter testaccel times than the real:
set testaccel 360

# Refresh period in seconds:
set refresh_period 1

# -------------- COMPUTABLE global variables
# Data: empty here
set data "-"

# UT string "YYYY MM DD hh mm ss"
set ut "0000 00 00 00 00 00"

# Julian day:
set jd 0

# Stellar time in HH MM SS.ss
set st 0

# Stellar time in HH.HHHHH
set sth 0

# Sun horizontal coordinates:
set hsun 0

# Moon horizontal coordinates:
set hmoon 0

# Sun and moon equatorial coordinates: RA in hours, DEC in degrees:
set sunra 0; set sundec 0; set moonra 0; set moondec 0

#
# Get current UTC using local machine time and system timezone set
#
# @param mode may be either "date", "time" or "full"
# @return UT as a string YYYY MM DD for mode==date,  YYYY MM DD hh mm ss for 
# mode==full or hh mm ss for mode==time or anything else
# ATTENTION! Machine local time and - ESPECIALLY - timezone should be correct!
#
proc get_ut {{mode full}} {
	global realmode startclock testaccel
	if {$realmode} {
		set cl [clock seconds]
	} else {
		set cl [expr $startclock + ([clock seconds] - $startclock)*$testaccel]
	}
	switch -glob $mode {
		f* {
			return [ \
	clock format $cl -format "%Y %m %d %H %M %S" -gmt true ]
			}
		d* {
			return [ \
	clock format $cl -format "%Y %m %d" -gmt true]
			}
		default {
			return [ \
	clock format $cl -format "%H %M %S" -gmt true ]
			}
	}		
}


# Last data update time in seconds-1969:
set refresh_time 0

# Recompute all circumstance parameters:
#
proc compute_circ {} {
	global refresh_time refresh_period ut jd st hsun longitude latitude \
		sunra sundec moonra moondec sth long lat hmoon
	# If data are still fresh - return:
	if {[clock seconds]-$refresh_time<$refresh_period} return
	# compute times:
	set ut [get_ut full]
	set time [get_ut time]
	set hours [xms2x $time]
	set jd [expr [jd $ut] + $hours/24.]
	set st [x2xms [set sth [st $ut $hours $long]]]

	# Compute Sun coordinates:
	set suncoord [mini_sun $jd]
	set sunra [lindex $suncoord 0];	set sundec [lindex $suncoord 1]
	set h [expr $sth - $sunra]

	set hsun [lindex [hadec2azalt $h $sundec $lat] 1]

	# Compute Moon coordinates:	
	set moon [mini_moon $jd]
	set moonra [lindex $moon 0];	set moondec [lindex $moon 1];
	set h [expr $sth - $moonra]

	set hmoon [lindex [hadec2azalt $h $moondec $lat] 1]

	set refresh_time [clock seconds]
	
	# Cut the non-significant digints in Sun and Moon coordinates:
	# Precision of coordinates is about 1arcmin, so 3 digits is enough.
	# This corresponds to 0.015 degree = 0.9' in right ascentions.
	foreach c {sundec moondec hsun hmoon} {
		upvar 0 $c val;	set val [format %.2f $val]	}
	foreach c {sunra moonra} {
		upvar 0 $c val;	set val [format %.3f $val]	}
}

proc cmd_get pars {
	# global array of parameter values cmp_pvar(param)=value and comp.status:
	global cmp_pvar cmp_status
	if {$pars!="ident"} compute_circ
	# here we allow to retrieve all global variables:
	foreach p $pars {
		if {[info exists ::$p]} {
			upvar #0 $p par; set cmp_pvar($p) $par 
		} else {
			set cmp_status ERSYN; break
		}
	}
}

# setting the parameters from list p1 v1 p2 v2... sent by name for upvar ref.
proc cmd_set pv_listname {
	global cmp_status longitude latitude realmode refresh_time par_set
	global startclock long lat
	upvar $pv_listname pv_list
	array set pvar $pv_list
	# - would be enough if pvar is made global, trusting input completely. 
	foreach pname [array names pvar] {
		switch $pname {
			"longitude"
				{
					# Check range:
					set bad [catch {set x [xms2x $pvar($pname)]}]
					if { $bad || $x<-12. || $x>+12.} { 
						set cmp_status ERANG
						break
					} else {
						set longitude $pvar($pname) 
						set long $x
						set refresh_time 0
					}
				}
			"latitude"
				{
					# Check range:
					set bad [catch {set x [xms2x $pvar($pname)]}]
					if { $bad || $x<-90 || $x>90} { 
						set cmp_status ERANG
						break
					} else {
						set latitude $pvar($pname) 
						set lat $x
						set refresh_time 0
					}
				}
			"mode"
				{
					set refresh_time 0
					if {[string tolower $pvar($pname)]=="real"} {
						set realmode 1
					} elseif {[string tolower $pvar($pname)]=="test"} {
						set realmode 0
						set startclock [clock seconds]
					} else {
						set cmp_status ERANG
						break
					}
				}
			default
				{
					puts "pname=$pname, parset=$par_set"
					if {[string match -nocase *$pname* $par_set]} {
						upvar #0 $pname par
						set par $pvar($pname)
					} else {
						set cmp_status ERSYN
						break
					}
				}
		}
	}
}

proc cmd_init {} {
	global cmp_status
	# read CFG file?
	set cmp_status READY
}

proc cmd_go_local {} {
	global cmp_status
	# enable accepting local console commands
	set cmp_status LOCAL
}

proc cmd_park {} {
	global cmp_status cmp_pvar
	# do nothing?
	set cmp_status PARKED
}

proc cmd_go_online {} {
	global cmp_status
	# disable local console commands
	set cmp_status READY
}

#######################################################################
########## Section of checking examples taken from AFK examples #######

proc examples_jd {} {
	# outputs from jd{} in different modes:
	puts [get_ut date];	puts [get_ut time];	puts [get_ut full]
	puts "Examples of JD:"
	foreach date [list "1957 10 4.81" "333 01 27.5" "-584 05 28.63"\
		"1900 01 01" "2000 01 01" [get_ut date]] {
		puts "JD($date) = [jd $date]"
	}	
	puts ""
}

proc examples_coord {} {
	global longitude
	puts "Examples of x2xms<->xms2x, deg2hms"
	foreach crd1 [list "50 47 55" "359 59 59.999" "362 10 20" "-15 00 00"\
		"04 34 00" "-59 59 59.999" "-00 17 25.94" "-15 -09 00"\
		"00 00 -30" "$longitude"] {
		puts -nonewline "xms2x($crd1) = [set x [xms2x $crd1]]"
		puts -nonewline " = [x2xms $x]"
		puts " = [deg2hms $x] h m s"
	}
}

proc examples_minisun {} {
	puts "Examples for minisun:"
	foreach date {"1978 11 12"} {
		puts "Sun coords for $date are: [set c [mini_sun [jd $date]]] h d"
		puts "RA in H M S: [x2xms [lindex $c 0]]"
		puts "DEC in D M S: [x2xms [lindex $c 1]]"
	}
	puts ""
}

proc examples_minimoon {} {
	puts "Examples for minimoon:"
	foreach date {"1979 12 07"} {
		puts "Moon coords for $date are: [set c [mini_moon [jd $date]]] h d"
		puts "RA in H M S: [x2xms [lindex $c 0]]"
		puts "DEC in D M S: [x2xms [lindex $c 1]]"
	}
	puts ""
}

proc examples_coords_trans {} {
	set ukkle [list "+0 17 25.94" "+50 47 55"]
	puts "Longitude Ukkle = [lindex $ukkle 0] = [set long \
		[xms2x [lindex $ukkle 0]]] hour"
	puts "Latitude Ukkle =[lindex $ukkle 1] = [set lat \
		[ xms2x [lindex $ukkle 1]]] degree"
	set saturn [list "10 57 35.681" "8 25 58.10"]
	puts "Saturn RA =[lindex $saturn 0] = [set ra \
		[xms2x [lindex $saturn 0]]] hour"
	puts "Saturn Dec =[lindex $saturn 1] = [set dec \
		[xms2x [lindex $saturn 1]]] degree"
	set date "1978 11 13"
	set time "04 34 00"
	set hours [xms2x $time]
	set st [st $date $hours $long]
	puts "St=$st hours = [x2xms $st] h m s (St0=[x2xms [st $date $hours 0]])"
	set h [expr $st - $ra]
	puts "HA(Saturn) = $h h = [x2xms $h] h m s = [expr $h*15] deg"
	puts "Az h = [hadec2azalt $h $dec $lat]"
	puts ""
}

proc examples_sunazalt {} {
	global longitude latitude
	set long [xms2x $longitude]; set lat [xms2x $latitude]
	set date [get_ut date]
	puts "Example of Sun az.alt: date $date, long=$longitude, lat=$latitude"
	puts "Time		Az\[deg\] 		Altitude\[deg\]"
	foreach hours { 12 13 14 15 16 17 18 19 20 } {
		# here we implicitly compute jd{} twice: in st{} and here for mini_sun{}
		set st [st $date $hours $long]
		set jd [expr [jd $date] + $hours/24.0]
		set suncoord [mini_sun $jd]
		set sunra [lindex $suncoord 0];	set sundec [lindex $suncoord 1]
		set h [expr $st - $sunra]
		puts "[x2xms $hours] [hadec2azalt $h $sundec $lat]"
	}
	puts ""
}

proc examples_all {} {
	examples_jd
	examples_coord
	examples_coords_trans
	examples_sunazalt
	examples_minisun
	examples_minimoon
}

open_server $port
#examples_all
exit
