#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

source compcmd.tcl


# DIMM interface simulator program.
# Port number:
set port 1235
# Identification:
set ident "dimm.tcl v0.02"

set time_of_data ""
set seeing 0
#
#... Return the local time stamp according to the adopted format:
#
set ut_format "%Y-%m-%d %H:%M:%S"
proc time_stamp {{sec1969 -1} {gmt true}} {
	global ut_format
	if {$sec1969==-1} {
		set sec1969 [clock seconds]
	}
	return [clock format $sec1969 -format $ut_format -gmt $gmt]
}

# Sumulator of DIMM data
proc refresh_data {} {
	global seeing time_of_data
	if {$seeing==0} {set seeing 0.5}
	set seeing [expr $seeing + rand()/2-0.25]
	if {$seeing<0.25} {set seeing 0.3} elseif {$seeing>3.0} {set seeing 1.5}
	set time_of_data [time_stamp]
}

# Compose a string with latest obtained results
proc get_data {} {
	global time_of_data seeing
	set s "ut=\\\"$time_of_data\\\" seeing=[format %4.2f $seeing]"
}

set reg_mom [clock seconds]
proc cmd_get pars {
	# global array of parameter values cmp_pvar(param)=value and comp.status:
	global cmp_pvar cmp_status ident reg_mom
	foreach p $pars {
		switch $p {
			"ident"
				{ set cmp_pvar($p) $ident }
			"data"
				{ set cmp_pvar($p) [get_data] }
			default
				{ set cmp_status ERSYN; break }
			"dra" - "ddec" 
				{
					# Simulate the deviations of a star from center:
					set range [expr [clock seconds] - $reg_mom]
					if {$range>100} {set range 100}
					set val [expr (2*rand()-1)*$range]
					if {$p=="dra"} {set reg_mom [clock seconds]}
					# Once in 30 cases we miss the star...:
					if {[expr int(rand()*100)]==1} {
						set cmp_status NOSTAR
						after 3000 {
							if {$cmp_status=="NOSTAR"} {set cmp_status READY}
						}
					}
					if {$cmp_status=="NOSTAR"} {
						set cmp_pvar(dra) 0
						set cmp_pvar(ddec) 0
					} else {
						set cmp_pvar($p) $val
					}
				}
		}
	}
}

# Status which must be restored by STOP NOW put here by INIT,PARK,RUN
set previous_status $cmp_status

proc cmd_init {} {
	global cmp_status cmp_pvar after_id previous_status
	# preparation time in seconds:
	set t_init 5
	# read CFG file?
	# simulate hardware preparation:
	set previous_status $cmp_status
	set cmp_status BUSY
	set cmp_pvar(wait) $t_init
	set after_id [after [expr $t_init*1000] {
		set cmp_status READY; set after_id 0
	}]
}

proc cmd_go_local {} {
	global cmp_status
	# enable accepting local console commands
	set cmp_status LOCAL
}

proc cmd_park {} {
	global cmp_status cmp_pvar after_id previous_status
	# simulate hardware shutdown
	set t_park 17
	set previous_status $cmp_status
	set cmp_status BUSY
	set cmp_pvar(wait) $t_park
	set after_id [after [expr $t_park*1000] {
		set cmp_status PARKED; set after_id 0
	}]
}

#
# Run command model: here we determine the needed time to run job
# and start the idle process of making the job itself. That job
# sets status when ready - this is detected by process_command in compcmd.tcl.
# While running job, the script must "update" the idle calls from time to time.
# because otherwise it cannot report status while being BUSY.
#
# timer identifier of run command: to be able to simulate its stop correctly
set after_id 0
proc cmd_run arglistname {
	global cmp_status cmp_pvar after_id previous_status
	upvar $arglistname arglist
	# status is set BUSY in command service...
	set previous_status READY
	set cmp_status BUSY
	set t_bkgr 5; set t_star 15
	if {[string match -nocase "*bkgr*" $arglist]} {
		puts "Measure background during $t_bkgr sec..."
		set t_job $t_bkgr
	} else {
		puts "Measure seeing during $t_star sec..." 
		set t_job $t_star
	}
	set cmp_pvar(wait) $t_job
	after idle cmd_run_idle $t_job
}

# Job-maker started on background in cmd_run.
proc cmd_run_idle t_job {
	global cmp_status after_id
	puts "start idle..."
	set c [clock seconds]
	set after_id 1
	while {[expr [clock seconds] - $c]<$t_job} { 
		# simulate the peace of job:
		after 100
		# update the idle calls:
		update
		# exit if STOP NOW commanded: 
		puts "running"
		if {$after_id!="1"} break
	}
	set after_id 0
	# simulate data:
	refresh_data
	puts "done idle";set cmp_status READY
}

proc cmd_stop args {
	global after_id cmp_status previous_status
	# stop accumulation of data under way.........
	# simulate it:
	if {![string match after* $after_id]} {
		# No after-script started: it's RUN:
		set after_id 2
#		while {$after_id==2} {update} - no sense to do: this proc must complete!
		puts "RUN stopping..."
	} else {
		# After-script was canceled: INIT
		puts "Long action (INIT/PARK) stopping"
	}
	puts "setting status back to $previous_status"
	set cmp_status $previous_status
}

# object set in DIMM
set object_hr 0
# setting the parameters from list p1 v1 p2 v2... sent by name for upvar ref.
proc cmd_set pv_listname {
	global cmp_status object_hr
	upvar $pv_listname pv_list
	array set pvar $pv_list
	# - would be enough if pvar is made global, trusting input completely. 
	foreach pname [array names pvar] {
		switch $pname {
			"object"
				{
					set hr $pvar($pname)
					puts -nonewline "Got object=$hr  "
					# simulation of checking object and identifying it in
					# local catalog:
					if {[string is digit $hr] && $hr<9876} {
						# HR ok
						set object_hr $hr
						puts "Set HR=$object_hr"
					} else {
						puts "Bad object ID"
						set cmp_status ERANG
						break
					}
				}
			default
				{
					set cmp_status ERSYN
					break
				}
		}
	}
}

proc cmd_go_online {} {
	global cmp_status
	# wait until local job is completed here:
	#............. vwait ...
	# disable local console commands.........
	set cmp_status READY
}

open_server $port
#while 1 { after 1000; puts "hsun=[format %-3.1f [compute_hsun]] deg" }
exit
