#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

# OBJM object manager program.

#
# USAGE: additional commands to CIRC and standard ones:
# SET STAR_CAT=<file name> - to set the prepared star list file name
# SET STAR_COLS="<table columns order>" - declare the structure of the table
#    of targets
# SET STAR_SEL="<selection criterion on column names: less-better>"
# SET STAR_REJ="<rejection criterion on columns: <>0 - reject>"
# RUN [HOUR] [ZEN] [MOON] <and/or any star table column names>] -
#    start sorting the table searching for the best star accroding to STAR_SEL 
#    and STAR_REJ set and return the value(s) of mentioned parameter(s):
#    HOUR: in decimal hours in (-12,+12]
#    ZEN: zenith distance in degrees
#    MOON: distance to Moon in degrees
#    other parameters measure and mean the same as ones in the input STAR_CAT.
# GET <same parameter names> - retrieve (other) parameters of resulting star
# GET HOUR ZEN - retrieve CURRENT (modified with time) hour angle and/or zen.d.
# GET STAR_COLS STAR_SEL STAR_REJ - retrieve the current settings
#

# Idea:
# Create the star ASCII list with coodinates represented as floating point 
# numbers RA and DEC where all potential targets for ANY site are given.
# This must be done manually, before starting to work with objm.tcl.
# Then the run-time selection is made by calls of the "exec" command to use the
# standard AWK and SORT utilities and read their output.
#
# With the prepared list, provided as a file star.lst in current directory, 
# the following steps of processing are performed each time the new star is 
# needed:
# 1. compute several new quantities (with AWK):
#     - hour angles in hours, 
#     - zenith distances in degrees and
#     - distance to Moon in degrees.
#   These quantities are printed as 1st several columns of the output table;
# the rest input columns are printed to the right.
# 
# 2. apply the REJECTION criterion: the logical expression allowed by AWK which 
#    is NON-ZERO for stars which CANNOT be targets in current moment. This means
#    that this expression is of the "<...> OR <...> OR ..." kind. The lines
#    which result in zero criterion value are printed "as is"
#
# 3. apply BEST star criterion: the mathematical expression allowed by AWK which
#    value is SMALLEST for best-suitable stars. This value is printed as a first
#    column of the output table; the rest of the output (the selected subtable
#    containing the newly computed columns made in stage.1) is printed to the 
#    right.
#
# 4. use the GNU (!!!) SORT program for sorting the resulting table according
#    to the NUMERICAL value of the first column. This is done with help of "-g"
#    option: "sort -g".
#
#  The first several lines are displayed on the console for checking purposes;
#  the values from the first line (that of the "best" star) are extracted
#  using the table disposition information (see cn_ra, cn_vmag, etc values which
#  are the numbers of the ra, dec... etc columns in the INPUT star list).
# 

# This program is an addition to the service of Circumstance monitor since
# we need astronomical data such as Moon coordinates, stellar time and utilities
# for representation in different forms. So first we source the stuff of 
# circ.tcl except for open_server and exit commands:
exec grep -v -e open_server -e ^\ *exit <circ.tcl >tmp.tmp
source tmp.tmp
exec rm tmp.tmp

# Overwrite the circ.tcl parameters:
# Port number:
set port 1238
# Identification:
set ident "objm.tcl v0.1"

# Star catalogue file name:
set star_cat star.lst

# Names of parameters which are listed in star list and which may be
# accessed with GET-commands. FIRST FOUR parameters are derived from the
# others during list processing, so the 1st column in the list has a meaning
# of the 4th element in star_cols.
#
# ATTENTION! Currently there is a restriction: no name of the column
# may be the part of the name of the other column (substitutions in rejection
# and selection criteria)
set star_cols "object ra dec vmag spcl"

# Additional derivative parameters computed in cmd_run and inserted in the left 
# part of the stellar data table before applying the criteria:
set dernames "sel hour zen moon"
# ATTENTION!!! These names cannot be modified without script edition, since they
# are used explicitly (search for "hour" and "zen")

# Number of derived star parameters:
set n_der [expr [llength $dernames] - 1]
# NOTE that the "sel"ection parameter is NOT included in n_der!!!!
# ALSO note, that addition of some parameter in dernames while improving the
# program in future INPLIES that the programmer adds computation and printing
# of this parameter in the first AWK processing script!

# Star catalogue structure: column numbers follow the order of columns 
# listed in "star_cols" variable (SET star_cols=... from SV to modify order!)
set i -$n_der
foreach p $dernames {set cn_$p $i; incr i}
foreach p $star_cols {set cn_$p $i; incr i}
unset i

# Criteria for rejection and selection the best star: operate (case-ignoring)
# with name-refernces to columns listed in "dernames" and "star_cols". These are
# the AWK expressions: star_rej is logical one and sbstitutes in processing as
# AWK script "if (!($star_rej)) {print}";
# star_sel is arithmetical expression which computes the selection value in AWK:
#  "$1 = $star_sel; print" which output is fed to sorting.
#
# Rejection criterion: hour angle beyond +-3 hours, zenith>80deg, 
# moon distance < 30deg, (declination-latitude)<5
set star_rej \
"(HOUR<-4)||(Hour>2) || (Zen>45)||(MOON<30)||((DEC<-25)&&(DEC>-35))||(VMAG>3)"
# Selection criterion: no abs() function in AWK!:
set star_sel "(HOUR+3)*(1-2*((HOUR+3)<0))+VMAG/10"

# Parameters allowed for setting by cmd_set (in circ.tcl) without checking:
set par_set "$par_set star_cols star_rej star_sel star_cat"
# Here we append these three parameters to par_set defined in circ.tcl.

# Length of the displayed table of best stars:
set tbest_len 10

# Process identifiers of AWK, sort etc. programs started on background for 
# processing the tables. Used to control the output-ready fact (by dissipation
# of these PIDs from the process list (ps -p) AND for stopping the processing
# by STOP NOW (makes sence for thousands lines of tables when the processing 
# time becomes significant (~10-15sec for 50000 lines of star.lst)
set sort_pid 0

proc cmd_run arglistname {
	global cmp_status cmp_pvar after_id previous_status
	set previous_status $cmp_status
	set cmp_status BUSY
	# Maximized time of sorting the information in star catalogue, [sec]:
	set t_job 30
	set cmp_pvar(wait) $t_job
	after 1 cmd_run_idle $arglistname
}

# Refresh the hour angle and zenith distance of the currently set star
# which coordinates are in global variables ra, dec. Stellar time serves
# as an indicator of access to cmd_get{} in CIRC, so we simply trace it:
trace variable sth w refresh_ha_zen
proc refresh_ha_zen {name name2 op} {
	global hourh hour zen rah decd sth lat
	# If hour still not exists - return (before first RUN)
	if {[info globals rah]=="" || [info globals decd]=="" || \
		[info globals hour]==""} {
		puts "Cannot refresh hour, zen: no vars still"
		return
	}
	puts -nonewline "Refresh H=$hour Z=$zen... "
	set hourh [expr $sth-$rah]
	# Correct range to (-12,+12):
	if {$hourh<-12} {set hourh [expr $hourh+24]} elseif {$hourh>+12} {
		set hourh [expr $hourh-24]}
	# convert back to form accepted for output with GET-command:
	set hour [x2xms $hourh]
	set zen [expr 90 - [lindex [hadec2azalt $hourh $decd $lat] 1]]
	puts "Hnew=$hour, Znew=$zen"
}

# The job-maker started on background from cmd_run{}.
proc cmd_run_idle arglistname {
	global cmp_status cmp_pvar after_id star_rej star_sel star_cat sth\
		longitude latitude moondec moonra tbest_len ut star_cols dernames n_der \
		sort_pid st
		
	# Here we use GNU awk for working on tables to add columns:
	# - Hour angle in hours in range [-12,+12], H
	# - zenith distance in degrees, Z
	# - distance to Moon in degrees, M
		
	# Refresh stellar time and Moon coordinates:
	compute_circ
	# Organize a pipe:
	# 1. Compute the needed quantities: H Z M: H=S-alpha; 
	# zenith distance Z: Here we use the formula(8.6) (AFK, see circ.tcl) 
	# for the sine of altitude given fi, delta and H (sinh), produce cosh and
	# aderive z=90-atan2(sinh,cosh)*180/pi;
	# Angular distance to Moon M: use formula(9.1)(AFK,p.44) (body "2" is Moon)
	#
	# 2,3,4. Apply rejection and selection criteria and sort
	set pi 3.14159265358979323
	# Sine and cosine of latitude:
	set lat [expr [xms2x $latitude]*$pi/180]
	set sf [expr sin($lat)]; set cf [expr cos($lat)]

	puts [string repeat "=" 70]
	puts "Stellar time: $st = [format %6.3f $sth] h"
	puts "Latitude: [format %6.3f $lat] rad = $latitude; [expr $sf*$sf+$cf*$cf]==1"
	
	# Moon coordinates: sin(dec), cos(dec) and alpha radians:
	set m_sd [expr sin($moondec*$pi/180)]
	set m_cd [expr cos($moondec*$pi/180)]
	set m_ra [expr $moonra*$pi/12]
	puts "Moon: [format {%6.3f h = %6.3f rad, %6.2f d} $moonra $m_ra $moondec]"

	# Convert the column names into the AWK column references in the rejection
	# and selection criteria:
	# Concat the original and derivative names:
	set parnames "$dernames  $star_cols"
	# Create the map of conversion:
	foreach p $parnames {
		# for formulii of AWK below:
		global cn_$p 
		# for map:
		upvar #0 cn_$p cnpar
		# we shift the numbers by n_der+1 since our cn_* reflect the INPUT table
		lappend map $p \$[expr $cnpar + $n_der + 1]
	}
	puts "Map of column names to numbers conversion (final table!):\n $map"
	set reject [string map -nocase $map $star_rej]
	set select [string map -nocase $map $star_sel]
	puts "Rejection: $reject\nSelection: $select"

	# Make computations: on background.
	puts -nonewline "Sorting the catalogue... "
	flush stdout
	# IN ORDER to have the same column references in rejection and selection,
	# we reserve sel-field in first AWK script:
	exec rm -f tmp.sel
	set sort_pid \
	[exec awk -v OFMT=%#9.5f "\{pi=$pi; ra=\$$cn_ra*pi/12; d=\$$cn_dec*pi/180; \
		h=$sth*pi/12-ra; if (h>pi) {h=h-2*pi}; if (h<-pi) {h=h+2*pi}; \
		sinh = $sf*sin(d)+$cf*cos(d)*cos(h); \
		cosh = sqrt(1.0-sinh*sinh); z=90-180*atan2(sinh,cosh)/pi; \
		cosm = sin(d)*$m_sd + cos(d)*$m_cd*cos(ra-$m_ra); \
		sinm = sqrt(1.0-cosm*cosm); m=180*atan2(sinm,cosm)/pi; \
		print 0,h*12/pi,z,m,\$0\}" $star_cat | \
		tee tmp.all | \
		\
		awk -v OFMT=%\#9.5f "\{if (!($reject)) print\}" | \
		\
		awk -v OFMT=%\#9.5f "\{\$1 = $select; print\}" | \
		sort -g | head -n $tbest_len > tmp.sel & ]

	# Wait untill the processes complete: we detect process completion by 
	# either catching an error or empty result of "ps -p <unexistent PID>".
	# (We are NOT SURE that all "ps" exit abnormally from listing absent PIDs)
	set dead 0
	while 1 {
		foreach pid $sort_pid {
			set psl ""
			set absent1 [catch {set psl [exec -- ps -h -p $pid]}]
			set absent2 [expr ([string length $psl]==0)]
			set dead [expr $absent1 || $absent2]
			if {!$dead} {
				# at least this PID still alive: wait a bit more
				update
				after 100
				continue
			}
		}
		# neither of processes exist any more:
		break
	}
	puts -nonewline "Waiting for output... "
	flush stdout
	# wait while the system pushes the output to the file:
	while {[lindex [set l [exec ls -s tmp.sel]] 0]==0} update
	puts "Output ready."
	set best_list [exec cat tmp.sel]

	# Best star string is the first line in best_list:
	set best_star [string range $best_list 0 \
		[expr [string first \n $best_list] - 1]]

	# Case of empty selection:
	if {[string length $best_star]==0} {
		set cmp_status NOSTAR
		return
	}
	
	# Display the best-stars table:
	puts "Best stars for $ut"
	set header ""
	foreach p $parnames { set header "$header$p\t"}
	puts $header
	puts [string repeat "-" 70]
	puts $best_list
	puts [string repeat "-" 70]

	# Export the parameters of this star
	set bs [split $best_star]
	set i 0
	foreach p $parnames {
		upvar #0 $p par
		set par [lindex $bs $i]
		incr i
	}
	# Produce the visibility time: CURRENTLY it is the time 
	# during which the abs(hour angle) of the star is not more than now:
	global tvis hour
	set tvis [expr int((abs($hour)-$hour)*3600/1.00273)]
	
	# Convert the angular parameters saving first decimals:
	global rah ra decd dec
	set rah $ra
	set decd $dec
	foreach p {ra dec hour} {
		if {[info globals $p]!=""} {
			upvar #0 $p par
			set par [x2xms $par]
		}
	}
	# Return the requested parameters: as in cmd_get (see circ.tcl)
	upvar #0 $arglistname args
	foreach p $args {
		if {$p!="switch" && [info globals $p]!=""} {
			upvar #0 $p par; set cmp_pvar($p) $par 
		}
	}
	set cmp_status READY
#	exec rm tmp.sel
}

proc cmd_stop args {
	global sort_pid cmp_status
	# stop sorting the objects:
	puts -nonewline "Stopping sorting..."
	foreach pid $sort_pid {
		catch {exec kill -9 $pid}
	}
	set cmp_status READY
	set sort_pid 0
	puts "Stopped"
}


open_server $port
exit
