#!/bin/sh
# the next line restarts using tclsh\
exec tclsh "$0" "$@"

source compcmd.tcl


# TLSP interface program (LX200 simulator)
# Port number:
set port 1239
# Identification:
set ident "tlsp.tcl v0.01"

# Simulation or real telescope:
set realmode true

# Source astronomical formulae:
source astro.tcl

# Default geographic coordinates: Moscow
# ATTENTION!!!!!! Longitude is positive to the EAST from Greenwich!!!!!!!!!
set long [xms2x "2 30 48"]; set lat [xms2x "55 45 00"]

# Value for unknown coordinates:
set undefined -99

# Preset coordinates for trials: commands GET TVIS TVIS2:
# RA: in hours, DEC: in degrees
set setra $undefined; set setdec $undefined

# Time of visibility in current and opposite orientations [seconds] for
# current coordinates:
set tvis_cur [list 0 0]

# Time of visibility in current and opposite orientations [seconds] for
# pre-set coordinates:
set tvis_set [list 0 0]

# Current coordinates of pointing and perekladka (orientation): +1 eastern,
# -1 western, 0 - unknown
# RA: [hours], DEC: [degrees]
set curra $undefined; set curdec $undefined; set curorient 0

# current stellar time:
set curst $undefined

# Reading the current position from the mount:
proc get_cur_coords {} {
	# ... reading RA and DEC into global curra, curdec
	
}

# Simulate the drive:
# Rate in hours per second:
set drive_rate 1.00273*1/3600.
set drive 0
proc stop_drive {} { global drive;	set drive 0}
proc start_drive {} { global drive drive_rate;	set drive $drive_rate;
simulate_drive}

# Check the TVIS for current coordinates. This will stop TLSP when limit is
# reached.
proc simulate_drive {} {
	global drive curra
	if {$drive>0} {
		update_coords
		after 5000 simulate_drive
	}
}

# Get the limits in hour angle given the declination and orientation
# @param dec declination in degree
# @param orient +1 - eastern, -1 - western 
# @return [list hmin hmax] in hours
#
# Note: for fork-type mounts, declination is limited to >decmin for all hour
# angles; western orientation is when the wires go out from that wing of the
# fork which faces East and vice versa.
# If no part of sphere is accessible given such a declination or orientation
# a list of { 0 0 } is returned.
#
set lim_dec_min -40
set lim_dec_max 90
# hour limits in eastern orientation: in western they are opposite in sign
set lim_hour_min -6
set lim_hour_max 12
proc get_h_lims {dec orient} {
	global lim_dec_min lim_dec_max lim_hour_min lim_hour_max lat
	set dmin $lim_dec_min
	set dmax $lim_dec_max
	if {$lat<0} {
		set dmin [expr (-1)*$lim_dec_max]
		set dmax [expr (-1)*$lim_dec_min]
	}
	if {$dec>=$dmin && $dec<=$dmax && [expr abs($orient)]==1} {
		if {$orient>0} { 
			return [list $lim_hour_min $lim_hour_max]
		} else {
			return [list [expr -$lim_hour_max] [expr -$lim_hour_min]]
		}
	} else {return [list 0 0]}
}

# Moment at which the update of tvis, tvis2, ra, dec variables is made
set updatetime 0
set refresh_period 1
# Change the status to ERANG warn_time before limit is reached:
set warn_time 60

# Update the current coordinates and recompute the visibility time in current
# and opposite orientation. Check if TVIS<60 sec, then switch the status to
# ERANG. If TVIS<2 sec, stop tracking. If pre-set coordinates are known, compute
# the TVIS, TVIS2 for preset coordinates.
proc update_coords {{idle 1}} {
	global updatetime refresh_period long lat curra curdec undefined
	global tvis_cur tvis_set curorient warn_time cmp_status drive curst
	if {[expr [clock seconds] - $updatetime] < $refresh_period} return
	# Get current RA and DEC of the tube:
	get_cur_coords
	# Current time:
	set secs [clock seconds]
	set ut 	[clock format $secs -format "%H %M %S" -gmt true ]
	set h [xms2x $ut]
	set date [clock format $secs -format "%Y %m %d" -gmt true]
	# Current stellar time [0..24):
	set curst [st $date $h $long]

	#Check hour angle limits for current and set coordinates and orientation:
puts "St = $curst hours. Curorient=$curorient"
	foreach type {cur set} {
		upvar #0 ${type}ra ra
		upvar #0 ${type}dec dec
	
		# Compute hour angle (-24..24):
		set ha [expr $curst - $ra]
		# Put curhour in a range [-12..12]:
		if {$ha<-12} {
			set ha [expr $ha+24]
		} elseif {$ha>12} {
			set ha [expr $ha-24]
		}
puts "$type settings: ra=$ra dec=$dec ha=$ha"
		set tvis_$type [list]
		foreach ori_mul { +1 -1 } {
			set h_min_max [get_h_lims $dec [expr $ori_mul * $curorient]]
			if {$ha>[lindex $h_min_max 0] && \
				$ha<[lindex $h_min_max 1] } {
				lappend tvis_$type [expr ([lindex $h_min_max 1] - $ha)*3600]
			} else {
				lappend tvis_$type 0
			}
		}
upvar #0 tvis_$type tv
puts " -> tvis: ($tv) secs"
	}
	
	# If current time of visibility is small - change status and stop drive:
	if {$idle && [lindex $tvis_cur 0]<$warn_time && $drive>0 && 
		$cmp_status!="ERANG"} {
		puts "Limit is close! Change status to ERANG..."
		set cmp_status ERANG
		if {[lindex $tvis_cur 0]<=2} {
			stop_drive
		}
	}

	set updatetime [clock seconds]
}

proc cmd_get pars {
	# global array of parameter values cmp_pvar(param)=value and comp.status:
	global cmp_pvar cmp_status ident setra undefined
	foreach p $pars {
		switch $p {
			"ident"
				{ set cmp_pvar($p) $ident }
			"tvis" - "tvis2" 
				{ update_coords 0; 
				if {$setra==$undefined} {
					upvar #0 tvis_cur tv
				} else {
					upvar #0 tvis_set tv
				}
				if {$p=="tvis"} {
					set cmp_pvar($p) [format %.0f [lindex $tv 0]]
				} else {
					set cmp_pvar($p) [format %.0f [lindex $tv 1]]
				}
				}
			"ra" - "dec"
				{ update_coords 0; upvar #0 cur$p crd; 
					set cmp_pvar($p) [x2xms $crd] }
			"longitude" - "latitude" 
				{ if {$p=="longitude"} {upvar #0 long crd } else {
					upvar #0 lat crd}
					set cmp_pvar($p) [x2xms $crd] }
			default
				{ set cmp_status ERSYN; break }
		}
	}
}


# setting the parameters from list p1 v1 p2 v2... sent by name for upvar ref.
proc cmd_set pv_listname {
	global cmp_status setra setdec realmode updatetime long lat
	upvar $pv_listname pv_list
	array set pvar $pv_list
	# - would be enough if pvar is made global, trusting input completely. 
	foreach pname [array names pvar] {
		switch $pname {
			"ra"
				{
					# Check range:
					set bad [catch {set x [xms2x $pvar($pname)]}]
					if { $bad || $x<0. || $x>+24.} { 
						set cmp_status ERANG
						break
					} else {
						set setra $x 
						set updatetime 0
					}
				}
			"longitude"
				{
					# Check range:
					set bad [catch {set x [xms2x $pvar($pname)]}]
					if { $bad || $x<-12. || $x>+12.} { 
						set cmp_status ERANG
						break
					} else {
						set long $x
					}
				}
			"dec" - "latitude"
				{
					# Check range:
					set bad [catch {set x [xms2x $pvar($pname)]}]
					if { $bad || $x<-90 || $x>90} { 
						set cmp_status ERANG
						break
					} else {
						if {$pname=="dec"} {
							set setdec $x 
							set updatetime 0
						} else {
							set lat $x
						}
					}
				}
			"mode"
				{
					set refresh_time 0
					if {[string tolower $pvar($pname)]=="real"} {
						set realmode 1
					} elseif {[string tolower $pvar($pname)]=="test"} {
						set realmode 0
					} else {
						set cmp_status ERANG
						break
					}
				}
			default
				{
					set cmp_status ERSYN
					break
				}
		}
	}
}

# Status which must be restored by STOP NOW put here by INIT,PARK,RUN
set previous_status $cmp_status

proc cmd_init {} {
	global cmp_status cmp_pvar after_id previous_status curra curdec curorient
	global curst
	# preparation time in seconds:
	set t_init 1
	# simulate hardware preparation:
	set previous_status $cmp_status
	set cmp_status BUSY
	set cmp_pvar(wait) $t_init
	set after_id [after [expr $t_init*1000] {
		set cmp_status READY; set after_id 0
	}]
	
	get_cur_coords
	#!!!!!!!! Tempo:
	set curorient +1
	set curra 0; set curdec 90
	start_drive
	set curra $curst
}

proc cmd_go_local {} {
	global cmp_status
	# enable accepting local console commands
	set cmp_status LOCAL
}

proc cmd_park {} {
	global cmp_status cmp_pvar after_id previous_status undefined updatetime
	# simulate hardware shutdown
	stop_drive
	set t_park 17
	set previous_status $cmp_status
	set cmp_status BUSY
	set cmp_pvar(wait) $t_park
	set after_id [after [expr $t_park*1000] {
		set cmp_status PARKED; set after_id 0
	}]
	foreach par {curra curdec setra setdec} {
		global $par
		set $par $undefined
	}
	set updatetime 0
}

# Compute the movement time
# @param ra destination ra
# @param dec destination dec
# @param over with orientation change flag
# @return time in seconds
proc move_time {ra dec over} {
	global curra curdec
	set point_rate 15
	set point_settle 2
	# Compute the times for rotation:
	set t_rotate_ra [expr $ra - $curra + $over*12]
puts "Delta RA = $t_rotate_ra hours"
	if {$t_rotate_ra<-12} {set t_rotate_ra [expr $t_rotate_ra + 24]
	} elseif {$t_rotate_ra>+12} {set t_rotate_ra [expr $t_rotate_ra - 24]}
	set t_rotate_ra [expr abs($t_rotate_ra)*15/$point_rate + $point_settle]
		
	set t_rotate_dec [expr $dec - $curdec + $over*2*(90-$dec)]
puts "Delta DEC = $t_rotate_dec deg"
	set t_rotate_dec [expr abs($t_rotate_dec)/$point_rate + $point_settle]
	if {$t_rotate_dec>$t_rotate_ra} {
		set t_job [expr ceil($t_rotate_dec)]
	} else {
		set t_job [expr ceil($t_rotate_ra)]
	} 
	return $t_job
}

#
# Run command model: here we determine the needed time to run job
# and start the idle process of making the job itself. That job
# sets status when ready - this is detected by process_command in compcmd.tcl.
# While running job, the script must "update" the idle calls from time to time.
# because otherwise it cannot report status while being BUSY.
#
# timer identifier of run command: to be able to simulate its stop correctly
set after_id 0
proc cmd_run arglistname {
	flush stdout
	global cmp_status cmp_pvar after_id previous_status setra setdec undefined
	global tvis_set curorient curra curdec updatetime warn_time
	upvar $arglistname arglist
	# status is set BUSY in command service...
	set previous_status READY
	set cmp_status BUSY
	
	# Parse the input arguments:
	array set argarray $arglist
	if {[lsearch [array names argarray] ra]>=0 &&
		[lsearch [array names argarray] dec]>=0} {
puts -nonewline "pointing to the given coordinates:"
		set try [catch {
			set setra [xms2x $argarray(ra)]
			set setdec [xms2x $argarray(dec)]
		}]
		if {$try} {set setra $undefined; set setdec $undefined}
		set updatetime 0
	} elseif {[lsearch [array names argarray] dra]>=0 &&
			[lsearch [array names argarray] ddec]>=0} {
puts -nonewline "correction to the given coordinates: "
		set rad [expr 3.14159265358979324 / 180]
		set setra [expr $curra + $argarray(dra) / 15. /3600./ cos($curdec*$rad)]
		set setdec [expr $curdec + $argarray(ddec) / 3600.0]
		set updatetime 0
	} else {
puts -nonewline "pointing to the pre-set coordinates: "
	}
puts "$setra $setdec"

	if {[lsearch [array names argarray] over]>=0} {
		set ori_mul -1
		set over 1
	} else {
		set ori_mul +1
		set over 0
	}
	# Check the pointing coordinates visibility times:
	update_coords 0
	set point_time [move_time $setra $setdec $over]

	if { [lindex $tvis_set $over]<=[expr $point_time+$warn_time] } {
		# pointing is impossible:
		set cmp_pvar(wait) 0
		puts "No reason to point: point time is $point_time, tvis=\
		[lindex $tvis_set $over]"
		set cmp_status ERANG
		return
	} else {
		puts "Attention! we will point now!" 
		set cmp_pvar(wait) $point_time
		after idle cmd_run_idle $setra $setdec [expr $curorient*$ori_mul] \
			$point_time
	}
	# reset the pre-set coordinates:
	set setra $undefined; set setdec $undefined
}

# Job-maker started on background in cmd_run.
proc cmd_run_idle {ra dec orient t_job} {
	global cmp_status after_id curra curdec curorient
	puts "start pointing..."
	stop_drive
	set c [clock seconds]
	set after_id 1
	set startra $curra
	set startdec $curdec
	while {[set d [expr [clock seconds] - $c]]<$t_job} { 
		# simulate the peace of job:
		after 100
		# update the idle calls:
		update
		# exit if STOP NOW commanded: 
		# simulate data:
		set curra [expr $startra + ($ra-$startra)*$d/$t_job]
		set curdec [expr $startdec + ($dec-$startdec)*$d/$t_job]
		puts "ra=$curra dec=$curdec..."
		if {$after_id!="1"} break
	}
	if {$after_id=="1"} {
		# final precise alignment:
		set curra $ra
		set curdec $dec
	}
	set after_id 0
	set curorient $orient
	start_drive
	puts "pointing done";set cmp_status READY
}

proc cmd_stop args {
	global after_id cmp_status previous_status
	# stop accumulation of data under way.........
	# simulate it:
	if {![string match after* $after_id]} {
		# No after-script started: it's RUN:
		set after_id 2
#		while {$after_id==2} {update} - no sense to do: this proc must complete!
		puts "RUN stopping..."
	} else {
		# After-script was canceled: INIT
		puts "Long action (INIT/PARK) stopping"
	}
	puts "setting status back to $previous_status"
	set cmp_status $previous_status
}

proc cmd_go_online {} {
	global cmp_status
	# wait until local job is completed here:
	#............. vwait ...
	# disable local console commands.........
	set cmp_status READY
}

proc bgerror msg {
	global errorInfoq
	puts "BGERROR: $msg"
	puts "Error= $errorInfo"
}
open_server $port
exit
